﻿using System;
using Xunit;
using Xunit.Abstractions;
//using GraphViewTests = Terminal.Gui.Views.GraphViewTests;

// Alias Console to MockConsole so we don't accidentally use Console
using Console = Terminal.Gui.FakeConsole;
using NStack;
using Terminal.Gui;

namespace Terminal.Gui.TopLevelTests {
	public class WindowTests {
		readonly ITestOutputHelper output;

		public WindowTests (ITestOutputHelper output)
		{
			this.output = output;
		}

		[Fact]
		public void New_Initializes ()
		{
			// Parameterless
			var r = new Window ();
			Assert.NotNull (r);
			Assert.Equal (ustring.Empty, r.Title);
			Assert.Equal (LayoutStyle.Computed, r.LayoutStyle);
			Assert.Equal ("Window()({X=0,Y=0,Width=0,Height=0})", r.ToString ());
			Assert.True (r.CanFocus);
			Assert.False (r.HasFocus);
			Assert.Equal (new Rect (0, 0, 0, 0), r.Bounds);
			Assert.Equal (new Rect (0, 0, 0, 0), r.Frame);
			Assert.Null (r.Focused);
			Assert.NotNull (r.ColorScheme);
			Assert.Equal (Dim.Fill (0), r.Width);
			Assert.Equal (Dim.Fill (0), r.Height);
			Assert.Null (r.X);
			Assert.Null (r.Y);
			Assert.False (r.IsCurrentTop);
			Assert.Empty (r.Id);
			Assert.NotEmpty (r.Subviews);
			Assert.False (r.WantContinuousButtonPressed);
			Assert.False (r.WantMousePositionReports);
			Assert.Null (r.SuperView);
			Assert.Null (r.MostFocused);
			Assert.Equal (TextDirection.LeftRight_TopBottom, r.TextDirection);

			// Empty Rect
			r = new Window (Rect.Empty, "title");
			Assert.NotNull (r);
			Assert.Equal ("title", r.Title);
			Assert.Equal (LayoutStyle.Absolute, r.LayoutStyle);
			Assert.Equal ("Window()({X=0,Y=0,Width=0,Height=0})", r.ToString ());
			Assert.True (r.CanFocus);
			Assert.False (r.HasFocus);
			Assert.Equal (new Rect (0, 0, 0, 0), r.Bounds);
			Assert.Equal (new Rect (0, 0, 0, 0), r.Frame);
			Assert.Null (r.Focused);
			Assert.NotNull (r.ColorScheme);
			Assert.Null (r.Width);       // All view Dim are initialized now in the IsAdded setter,
			Assert.Null (r.Height);      // avoiding Dim errors.
			Assert.Null (r.X);           // All view Pos are initialized now in the IsAdded setter,
			Assert.Null (r.Y);           // avoiding Pos errors.
			Assert.False (r.IsCurrentTop);
			Assert.Empty (r.Id);
			Assert.NotEmpty (r.Subviews);
			Assert.False (r.WantContinuousButtonPressed);
			Assert.False (r.WantMousePositionReports);
			Assert.Null (r.SuperView);
			Assert.Null (r.MostFocused);
			Assert.Equal (TextDirection.LeftRight_TopBottom, r.TextDirection);

			// Rect with values
			r = new Window (new Rect (1, 2, 3, 4), "title");
			Assert.Equal ("title", r.Title);
			Assert.NotNull (r);
			Assert.Equal (LayoutStyle.Absolute, r.LayoutStyle);
			Assert.Equal ("Window()({X=1,Y=2,Width=3,Height=4})", r.ToString ());
			Assert.True (r.CanFocus);
			Assert.False (r.HasFocus);
			Assert.Equal (new Rect (0, 0, 3, 4), r.Bounds);
			Assert.Equal (new Rect (1, 2, 3, 4), r.Frame);
			Assert.Null (r.Focused);
			Assert.NotNull (r.ColorScheme);
			Assert.Null (r.Width);
			Assert.Null (r.Height);
			Assert.Null (r.X);
			Assert.Null (r.Y);
			Assert.False (r.IsCurrentTop);
			Assert.Empty (r.Id);
			Assert.NotEmpty (r.Subviews);
			Assert.False (r.WantContinuousButtonPressed);
			Assert.False (r.WantMousePositionReports);
			Assert.Null (r.SuperView);
			Assert.Null (r.MostFocused);
			Assert.Equal (TextDirection.LeftRight_TopBottom, r.TextDirection);
			r.Dispose ();
		}

		[Fact]
		public void Set_Title_Fires_TitleChanging ()
		{
			var r = new Window ();
			Assert.Equal (ustring.Empty, r.Title);

			string expectedOld = null;
			string expectedDuring = null;
			string expectedAfter = null;
			bool cancel = false;
			r.TitleChanging += (args) => {
				Assert.Equal (expectedOld, args.OldTitle);
				Assert.Equal (expectedDuring, args.NewTitle);
				args.Cancel = cancel;
			};

			expectedOld = string.Empty;
			r.Title = expectedDuring = expectedAfter = "title";
			Assert.Equal (expectedAfter, r.Title.ToString ());

			expectedOld = r.Title.ToString ();
			r.Title = expectedDuring = expectedAfter = "a different title";
			Assert.Equal (expectedAfter, r.Title.ToString ());

			// Now setup cancelling the change and change it back to "title"
			cancel = true;
			expectedOld = r.Title.ToString ();
			r.Title = expectedDuring = "title";
			Assert.Equal (expectedAfter, r.Title.ToString ());
			r.Dispose ();

		}

		[Fact]
		public void Set_Title_Fires_TitleChanged ()
		{
			var r = new Window ();
			Assert.Equal (ustring.Empty, r.Title);

			string expectedOld = null;
			string expected = null;
			r.TitleChanged += (args) => {
				Assert.Equal (expectedOld, args.OldTitle);
				Assert.Equal (r.Title, args.NewTitle);
			};

			expected = "title";
			expectedOld = r.Title.ToString ();
			r.Title = expected;
			Assert.Equal (expected, r.Title.ToString ());

			expected = "another title";
			expectedOld = r.Title.ToString ();
			r.Title = expected;
			Assert.Equal (expected, r.Title.ToString ());
			r.Dispose ();
		}

		[Fact, AutoInitShutdown]
		public void MenuBar_And_StatusBar_Inside_Window ()
		{
			var menu = new MenuBar (new MenuBarItem [] {
				new MenuBarItem ("File", new MenuItem [] {
					new MenuItem ("Open", "", null),
					new MenuItem ("Quit", "", null),
				}),
				new MenuBarItem ("Edit", new MenuItem [] {
					new MenuItem ("Copy", "", null),
				})
			});

			var sb = new StatusBar (new StatusItem [] {
				new StatusItem (Key.CtrlMask | Key.Q, "~^Q~ Quit", null),
				new StatusItem (Key.CtrlMask | Key.O, "~^O~ Open", null),
				new StatusItem (Key.CtrlMask | Key.C, "~^C~ Copy", null),
			});

			var fv = new FrameView ("Frame View") {
				Y = 1,
				Width = Dim.Fill (),
				Height = Dim.Fill (1)
			};
			var win = new Window ();
			win.Add (menu, sb, fv);
			var top = Application.Top;
			top.Add (win);
			Application.Begin (top);
			((FakeDriver)Application.Driver).SetBufferSize (20, 10);

			TestHelpers.AssertDriverContentsWithFrameAre (@"
┌──────────────────┐
│ File  Edit       │
│┌ Frame View ────┐│
││                ││
││                ││
││                ││
││                ││
│└────────────────┘│
│ ^Q Quit │ ^O Open│
└──────────────────┘", output);

			((FakeDriver)Application.Driver).SetBufferSize (40, 20);

			TestHelpers.AssertDriverContentsWithFrameAre (@"
┌──────────────────────────────────────┐
│ File  Edit                           │
│┌ Frame View ────────────────────────┐│
││                                    ││
││                                    ││
││                                    ││
││                                    ││
││                                    ││
││                                    ││
││                                    ││
││                                    ││
││                                    ││
││                                    ││
││                                    ││
││                                    ││
││                                    ││
││                                    ││
│└────────────────────────────────────┘│
│ ^Q Quit │ ^O Open │ ^C Copy          │
└──────────────────────────────────────┘", output);

			((FakeDriver)Application.Driver).SetBufferSize (20, 10);

			TestHelpers.AssertDriverContentsWithFrameAre (@"
┌──────────────────┐
│ File  Edit       │
│┌ Frame View ────┐│
││                ││
││                ││
││                ││
││                ││
│└────────────────┘│
│ ^Q Quit │ ^O Open│
└──────────────────┘", output);
		}

		[Fact, AutoInitShutdown]
		public void Activating_MenuBar_By_Alt_Key_Does_Not_Throw ()
		{
			var menu = new MenuBar (new MenuBarItem [] {
				new MenuBarItem ("Child", new MenuItem [] {
					new MenuItem ("_Create Child", "", null)
				})
			});
			var win = new Window ();
			win.Add (menu);
			Application.Top.Add (win);
			Application.Begin (Application.Top);

			var exception = Record.Exception (() => win.ProcessHotKey (new KeyEvent (Key.AltMask, new KeyModifiers { Alt = true })));
			Assert.Null (exception);
		}

		[Fact, AutoInitShutdown]
		public void Adding_Center_Window_Child_To_Window_Parent_Always_LayoutSubviews ()
		{
			var parentWin = new Window ();
			View childWin = null;

			var menu = new MenuBar (new MenuBarItem [] {
				new MenuBarItem ("Child", new MenuItem [] {
					new MenuItem ("_Create Child", "", () => {
						childWin = new Window () {
							X = Pos.Center (),
							Y = Pos.Center (),
							Width = 10,
							Height = 10
						};
						parentWin.Add (childWin);
					})
				}),
				new MenuBarItem ("View", new MenuBarItem [] {
					new MenuBarItem ("Create", new MenuItem [] {
						new MenuItem("_TextField", "", () => {
							var tf = new TextField ("Test") {
								X = Pos.Center (),
								Y = Pos.Center (),
								Width = 5
							};
							childWin.Add (tf);
						})
					})
				})
			});

			parentWin.Add (menu);
			Application.Top.Add (parentWin);
			Application.Begin (Application.Top);
			((FakeDriver)Application.Driver).SetBufferSize (20, 20);

			TestHelpers.AssertDriverContentsWithFrameAre (@"
┌──────────────────┐
│ Child  View      │
│                  │
│                  │
│                  │
│                  │
│                  │
│                  │
│                  │
│                  │
│                  │
│                  │
│                  │
│                  │
│                  │
│                  │
│                  │
│                  │
│                  │
└──────────────────┘", output);

			Assert.True (parentWin.ProcessHotKey (new KeyEvent (Key.C | Key.AltMask, new KeyModifiers { Alt = true })));
			Application.MainLoop.MainIteration ();
			parentWin.Redraw (parentWin.Bounds);
			TestHelpers.AssertDriverContentsWithFrameAre (@"
┌──────────────────┐
│ Child  View      │
│                  │
│                  │
│                  │
│    ┌────────┐    │
│    │        │    │
│    │        │    │
│    │        │    │
│    │        │    │
│    │        │    │
│    │        │    │
│    │        │    │
│    │        │    │
│    └────────┘    │
│                  │
│                  │
│                  │
│                  │
└──────────────────┘", output);

			Assert.True (parentWin.ProcessHotKey (new KeyEvent (Key.T | Key.AltMask, new KeyModifiers { Alt = true })));
			Application.MainLoop.MainIteration ();
			childWin.Redraw (childWin.Bounds);
			TestHelpers.AssertDriverContentsWithFrameAre (@"
┌──────────────────┐
│ Child  View      │
│                  │
│                  │
│                  │
│    ┌────────┐    │
│    │        │    │
│    │        │    │
│    │        │    │
│    │ Test   │    │
│    │        │    │
│    │        │    │
│    │        │    │
│    │        │    │
│    └────────┘    │
│                  │
│                  │
│                  │
│                  │
└──────────────────┘", output);
		}

		[Fact, AutoInitShutdown]
		public void Window_On_Non_Toplevel_Superview_Cannot_Overflows_His_Bounds ()
		{
			var win = new Window () { Width = Dim.Fill (5), Height = Dim.Fill (5) };
			var view = new View () { X = 3, Y = 3, Width = 10, Height = 10 };
			view.Add (win);
			Application.Top.Add (view);
			Application.Begin (Application.Top);

			ReflectionTools.InvokePrivate (
				typeof (Application),
				"ProcessMouseEvent",
				new MouseEvent () {
					X = 3,
					Y = 3,
					Flags = MouseFlags.Button1Pressed
				});

			Assert.Equal (win, Application.MouseGrabView);

			ReflectionTools.InvokePrivate (
				typeof (Application),
				"ProcessMouseEvent",
				new MouseEvent () {
					X = 2,
					Y = 2,
					Flags = MouseFlags.Button1Pressed | MouseFlags.ReportMousePosition
				});

			Assert.Equal (win, Application.MouseGrabView);
			Assert.Equal (new Rect (0, 0, 5, 5), win.Frame);

			ReflectionTools.InvokePrivate (
				typeof (Application),
				"ProcessMouseEvent",
				new MouseEvent () {
					X = 14,
					Y = 14,
					Flags = MouseFlags.Button1Pressed | MouseFlags.ReportMousePosition
				});

			Assert.Equal (win, Application.MouseGrabView);
			Assert.Equal (new Rect (4, 4, 5, 5), win.Frame);
		}
	}
}
